"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CommentsService = void 0;
const tslib_1 = require("tslib");
const debounce_1 = tslib_1.__importDefault(require("lodash/debounce"));
const flattenDeep_1 = tslib_1.__importDefault(require("lodash/flattenDeep"));
const groupBy_1 = tslib_1.__importDefault(require("lodash/groupBy"));
const di_1 = require("@opensumi/di");
const ide_core_browser_1 = require("@opensumi/ide-core-browser");
const browser_1 = require("@opensumi/ide-editor/lib/browser");
const ide_main_layout_1 = require("@opensumi/ide-main-layout");
const ide_monaco_1 = require("@opensumi/ide-monaco");
const browser_2 = require("@opensumi/ide-monaco/lib/browser");
const ide_theme_1 = require("@opensumi/ide-theme");
const textModel = tslib_1.__importStar(require("@opensumi/monaco-editor-core/esm/vs/editor/common/model/textModel"));
const common_1 = require("../common");
const comments_panel_view_1 = require("./comments-panel.view");
const comments_thread_1 = require("./comments-thread");
const tree_node_defined_1 = require("./tree/tree-node.defined");
let CommentsService = class CommentsService extends ide_core_browser_1.Disposable {
    constructor() {
        super(...arguments);
        this.decorationChangeEmitter = new ide_core_browser_1.Emitter();
        this.threads = new Map();
        this.threadsChangeEmitter = new ide_core_browser_1.Emitter();
        this.threadsCommentChangeEmitter = new ide_core_browser_1.Emitter();
        this.threadsCreatedEmitter = new ide_core_browser_1.Emitter();
        this.rangeProviderMap = new Map();
        this.rangeOwner = new Map();
        this.providerDecorationCache = new ide_core_browser_1.LRUCache(10000);
        // 默认在 file 协议和 git 协议中显示评论数据
        this.shouldShowCommentsSchemes = new Set(['file', 'git']);
        this.decorationProviderDisposer = ide_core_browser_1.Disposable.NULL;
    }
    get commentsThreads() {
        return [...this.threads.values()];
    }
    get isMultiCommentsForSingleLine() {
        var _a;
        return !!((_a = this.commentsFeatureRegistry.getConfig()) === null || _a === void 0 ? void 0 : _a.isMultiCommentsForSingleLine);
    }
    get currentAuthorAvatar() {
        var _a, _b;
        return (_b = (_a = this.commentsFeatureRegistry.getConfig()) === null || _a === void 0 ? void 0 : _a.author) === null || _b === void 0 ? void 0 : _b.avatar;
    }
    get filterThreadDecoration() {
        var _a;
        return (_a = this.commentsFeatureRegistry.getConfig()) === null || _a === void 0 ? void 0 : _a.filterThreadDecoration;
    }
    get onThreadsChanged() {
        return this.threadsChangeEmitter.event;
    }
    get onThreadsCommentChange() {
        return this.threadsCommentChangeEmitter.event;
    }
    get onThreadsCreated() {
        return this.threadsCreatedEmitter.event;
    }
    /**
     * -------------------------------- IMPORTANT --------------------------------
     * 需要注意区分 model.IModelDecorationOptions 与 monaco.editor.IModelDecorationOptions 两个类型
     * 将 model.IModelDecorationOptions 类型的对象传给签名为 monaco.editor.IModelDecorationOptions 的方法时需要做 Type Assertion
     * 这是因为 monaco.d.ts 与 vs/editor/common/model 分别导出了枚举 TrackedRangeStickiness
     * 这种情况下两个枚举的类型是不兼容的，即使他们是同一段代码的编译产物
     * -------------------------------- IMPORTANT --------------------------------
     * @param thread
     */
    createThreadDecoration(thread) {
        var _a;
        // 对于新增的空的 thread，默认显示当前用户的头像，否则使用第一个用户的头像
        const avatar = thread.comments.length === 0 ? this.currentAuthorAvatar : (_a = thread.comments[0].author.iconPath) === null || _a === void 0 ? void 0 : _a.toString();
        const icon = avatar
            ? this.iconService.fromIcon('', avatar, ide_theme_1.IconType.Background)
            : this.iconService.fromString('$(comment-unresolved)');
        const decorationOptions = {
            description: 'comments-thread-decoration',
            // 创建评论显示在 glyph margin 处
            glyphMarginClassName: ['comments-decoration', 'comments-thread', icon].join(' '),
        };
        return textModel.ModelDecorationOptions.createDynamic(decorationOptions);
    }
    createHoverDecoration() {
        const decorationOptions = {
            description: 'comments-hover-decoration',
            linesDecorationsClassName: ['comments-decoration', 'comments-add', (0, ide_core_browser_1.getIcon)('add-comments')].join(' '),
        };
        return textModel.ModelDecorationOptions.createDynamic(decorationOptions);
    }
    init() {
        const schemes = this.resourceService.getSupportedSchemes();
        for (const scheme of schemes) {
            this.shouldShowCommentsSchemes.add(scheme);
        }
        // 插件注册 ResourceProvider 时重新注册 CommentDecorationProvider
        // 例如 Github Pull Request 插件的 scheme 为 pr
        this.addDispose(this.resourceService.onRegisterResourceProvider((provider) => {
            if (provider.scheme) {
                this.shouldShowCommentsSchemes.add(provider.scheme);
                this.registerDecorationProvider();
            }
        }));
        this.addDispose(this.resourceService.onUnregisterResourceProvider((provider) => {
            if (provider.scheme) {
                this.shouldShowCommentsSchemes.delete(provider.scheme);
                this.registerDecorationProvider();
            }
        }));
        this.registerDecorationProvider();
    }
    handleOnCreateEditor(editor) {
        const disposer = new ide_core_browser_1.Disposable();
        disposer.addDispose(editor.monacoEditor.onMouseDown((event) => {
            if (event.target.type === browser_2.monacoBrowser.editor.MouseTargetType.GUTTER_LINE_DECORATIONS &&
                event.target.element &&
                event.target.element.className.indexOf('comments-add') > -1) {
                const { target } = event;
                if (target && target.range) {
                    const { range } = target;
                    // 如果已经存在一个待输入的评论组件，则不创建新的
                    if (!this.commentsThreads.some((thread) => thread.comments.length === 0 &&
                        thread.uri.isEqual(editor.currentUri) &&
                        thread.range.startLineNumber === range.startLineNumber)) {
                        const thread = this.createThread(editor.currentUri, range);
                        thread.show(editor);
                    }
                    event.event.stopPropagation();
                }
            }
            else if (event.target.type === browser_2.monacoBrowser.editor.MouseTargetType.GUTTER_GLYPH_MARGIN &&
                event.target.element &&
                event.target.element.className.indexOf('comments-thread') > -1) {
                const { target } = event;
                if (target && target.range) {
                    const { range } = target;
                    const threads = this.commentsThreads.filter((thread) => thread.uri.isEqual(editor.currentUri) && thread.range.startLineNumber === range.startLineNumber);
                    if (threads.length) {
                        // 判断当前 widget 是否是显示的
                        const isShowWidget = threads.some((thread) => thread.isShowWidget(editor));
                        if (isShowWidget) {
                            threads.forEach((thread) => thread.hide(editor));
                        }
                        else {
                            threads.forEach((thread) => thread.show(editor));
                        }
                    }
                }
                event.event.stopPropagation();
            }
        }));
        let oldDecorations = [];
        disposer.addDispose(editor.monacoEditor.onMouseMove((0, debounce_1.default)(async (event) => {
            const uri = editor.currentUri;
            const range = event.target.range;
            if (uri && range && (await this.shouldShowHoverDecoration(uri, range))) {
                oldDecorations = editor.monacoEditor.deltaDecorations(oldDecorations, [
                    {
                        range: (0, ide_monaco_1.positionToRange)(range.startLineNumber),
                        options: this.createHoverDecoration(),
                    },
                ]);
            }
            else {
                oldDecorations = editor.monacoEditor.deltaDecorations(oldDecorations, []);
            }
        }, 10)));
        disposer.addDispose(editor.monacoEditor.onMouseLeave((0, debounce_1.default)(() => {
            oldDecorations = editor.monacoEditor.deltaDecorations(oldDecorations, []);
        }, 10)));
        return disposer;
    }
    async shouldShowHoverDecoration(uri, range) {
        if (!this.shouldShowCommentsSchemes.has(uri.scheme)) {
            return false;
        }
        const contributionRanges = await this.getContributionRanges(uri);
        const isProviderRanges = contributionRanges.some((contributionRange) => range.startLineNumber >= contributionRange.startLineNumber &&
            range.startLineNumber <= contributionRange.endLineNumber);
        // 如果不支持对同一行进行多个评论，那么过滤掉当前有 thread 行号的 decoration
        const isShowHoverToSingleLine = this.isMultiCommentsForSingleLine ||
            !this.commentsThreads.some((thread) => thread.uri.isEqual(uri) && thread.range.startLineNumber === range.startLineNumber);
        return isProviderRanges && isShowHoverToSingleLine;
    }
    createThread(uri, range, options = {
        comments: [],
        readOnly: false,
    }) {
        // 获取当前 range 的 providerId，用于 commentController contextKey 的生成
        const providerId = this.getProviderIdsByLine(range.startLineNumber)[0];
        const thread = this.injector.get(comments_thread_1.CommentsThread, [uri, range, providerId, options]);
        thread.onDispose(() => {
            this.threads.delete(thread.id);
            this.threadsChangeEmitter.fire(thread);
            this.decorationChangeEmitter.fire(uri);
        });
        thread.onDidChange(() => {
            this.threadsChangeEmitter.fire(thread);
        });
        this.threads.set(thread.id, thread);
        this.addDispose(thread);
        this.threadsChangeEmitter.fire(thread);
        this.threadsCreatedEmitter.fire(thread);
        this.decorationChangeEmitter.fire(uri);
        return thread;
    }
    getThreadsByUri(uri) {
        return (this.commentsThreads
            .filter((thread) => thread.uri.isEqual(uri))
            // 默认按照 rang 顺序 升序排列
            .sort((a, b) => a.range.startLineNumber - b.range.startLineNumber));
    }
    handleCommentFileNode(parent) {
        const childs = [];
        const commentThreads = [...this.threads.values()].filter((thread) => thread.comments.length);
        const threadUris = (0, groupBy_1.default)(commentThreads, (thread) => thread.uri);
        Object.keys(threadUris).map((uri) => {
            var _a;
            const threads = threadUris[uri];
            if (threads.length === 0) {
                return;
            }
            const workspaceDir = new ide_core_browser_1.URI(this.appConfig.workspaceDir);
            const resource = new ide_core_browser_1.URI(uri);
            const description = (_a = workspaceDir.relative(resource)) === null || _a === void 0 ? void 0 : _a.toString();
            childs.push(new tree_node_defined_1.CommentFileNode(this, threads, description, resource.codeUri.fsPath, this.labelService.getIcon(resource), resource, parent));
        });
        return childs;
    }
    handleCommentContentNode(parent) {
        const childs = [];
        for (const thread of parent.threads) {
            const [first] = thread.comments;
            const comment = typeof first.body === 'string' ? first.body : first.body.value;
            childs.push(new tree_node_defined_1.CommentContentNode(this, thread, comment, `[Ln ${thread.range.startLineNumber}]`, first.author.iconPath
                ? this.iconService.fromIcon('', first.author.iconPath.toString(), ide_theme_1.IconType.Background)
                : (0, ide_core_browser_1.getIcon)('message'), first.author, parent.resource, parent));
        }
        return childs;
    }
    handleCommentReplyNode(parent) {
        const childs = [];
        const thread = parent.thread;
        const [_, ...others] = thread.comments;
        const lastReply = others[others.length - 1].author.name;
        childs.push(new tree_node_defined_1.CommentReplyNode(this, thread, (0, ide_core_browser_1.formatLocalize)('comment.reply.count', (others === null || others === void 0 ? void 0 : others.length) || 0), (0, ide_core_browser_1.formatLocalize)('comment.reply.lastReply', lastReply), '', parent.resource, parent));
        return childs;
    }
    async resolveChildren(parent) {
        let childs = [];
        if (!parent) {
            childs.push(new tree_node_defined_1.CommentRoot(this));
        }
        else {
            if (tree_node_defined_1.CommentRoot.isRoot(parent)) {
                childs = this.handleCommentFileNode(parent);
            }
            else if (tree_node_defined_1.CommentFileNode.is(parent)) {
                childs = this.handleCommentContentNode(parent);
            }
            else if (tree_node_defined_1.CommentContentNode.is(parent)) {
                childs = this.handleCommentReplyNode(parent);
            }
        }
        if (childs.length === 1 && tree_node_defined_1.CommentRoot.isRoot(childs[0])) {
            return childs;
        }
        const handlers = this.commentsFeatureRegistry.getCommentsPanelTreeNodeHandlers();
        if (handlers.length > 0) {
            for (const handler of handlers) {
                childs = handler(childs);
            }
        }
        return childs;
    }
    async getContributionRanges(uri) {
        // 一个diff editor对应两个uri，两个uri的rangeOwner不应该互相覆盖
        const cache = this.providerDecorationCache.get(uri.toString());
        // 优先从缓存中拿
        if (cache) {
            return await cache.promise;
        }
        const model = this.documentService.getModelReference(uri, 'get-contribution-rages');
        const rangePromise = [];
        for (const rangeProvider of this.rangeProviderMap) {
            const [id, provider] = rangeProvider;
            rangePromise.push((async () => {
                const ranges = await provider.getCommentingRanges(model === null || model === void 0 ? void 0 : model.instance);
                if (ranges && ranges.length) {
                    // FIXME: ranges 会被 Diff uri 的两个 range 互相覆盖，导致可能根据行查不到 provider
                    this.rangeOwner.set(id, ranges);
                }
                return ranges;
            })());
        }
        const deferredRes = new ide_core_browser_1.Deferred();
        this.providerDecorationCache.set(uri.toString(), deferredRes);
        const res = (await Promise.all(rangePromise)).filter(Boolean);
        // 消除 document 引用
        model === null || model === void 0 ? void 0 : model.dispose();
        // 拍平，去掉 undefined
        const flattenRange = (0, flattenDeep_1.default)(res).filter(Boolean);
        deferredRes.resolve(flattenRange);
        return flattenRange;
    }
    fireThreadCommentChange(thread) {
        this.threadsCommentChangeEmitter.fire(thread);
    }
    registerDecorationProvider() {
        // dispose 掉上一个 decorationProvider
        this.decorationProviderDisposer.dispose();
        this.decorationProviderDisposer = this.editorDecorationCollectionService.registerDecorationProvider({
            schemes: [...this.shouldShowCommentsSchemes.values()],
            key: 'comments',
            onDidDecorationChange: this.decorationChangeEmitter.event,
            provideEditorDecoration: (uri) => this.commentsThreads
                .map((thread) => {
                if (thread.uri.isEqual(uri)) {
                    if (thread.comments.length) {
                        // 存在评论内容 恢复之前的现场
                        thread.showWidgetsIfShowed();
                    }
                }
                else {
                    // 临时隐藏，当切回来时会恢复
                    thread.hideWidgetsByDispose();
                }
                return thread;
            })
                .filter((thread) => {
                const isCurrentThread = thread.uri.isEqual(uri);
                if (this.filterThreadDecoration) {
                    return isCurrentThread && this.filterThreadDecoration(thread);
                }
                return isCurrentThread;
            })
                .map((thread) => ({
                range: thread.range,
                options: this.createThreadDecoration(thread),
            })),
        });
        this.addDispose(this.decorationProviderDisposer);
    }
    registerCommentPanel() {
        // 面板只注册一次
        if (this.layoutService.getTabbarHandler(common_1.CommentPanelId)) {
            return;
        }
        this.layoutService.collectTabbarComponent([
            {
                id: common_1.CommentPanelId,
                component: comments_panel_view_1.CommentsPanel,
            },
        ], Object.assign({ badge: this.panelBadge, containerId: common_1.CommentPanelId, title: (0, ide_core_browser_1.localize)('comments').toUpperCase(), hidden: false, activateKeyBinding: 'ctrlcmd+shift+c' }, this.commentsFeatureRegistry.getCommentsPanelOptions()), 'bottom');
    }
    get panelBadge() {
        const length = this.commentsThreads.length;
        return length ? length + '' : '';
    }
    registerCommentRangeProvider(id, provider) {
        this.rangeProviderMap.set(id, provider);
        // 注册一个新的 range provider 后清理掉之前的缓存
        this.providerDecorationCache.clear();
        return ide_core_browser_1.Disposable.create(() => {
            this.rangeProviderMap.delete(id);
            this.rangeOwner.delete(id);
            this.providerDecorationCache.clear();
        });
    }
    forceUpdateDecoration() {
        var _a, _b, _c;
        // 默认适应当前 uri 去强刷 decoration
        // 这个值为 core editor 或者 modified editor
        const uri = (_a = this.workbenchEditorService.currentEditor) === null || _a === void 0 ? void 0 : _a.currentUri;
        uri && this.decorationChangeEmitter.fire(uri);
        // diffeditor 的 originalUri 也需要更新 Decoration
        const originalUri = (_c = (_b = this.workbenchEditorService.currentEditorGroup) === null || _b === void 0 ? void 0 : _b.diffEditor) === null || _c === void 0 ? void 0 : _c.originalEditor.currentUri;
        originalUri && this.decorationChangeEmitter.fire(originalUri);
    }
    getProviderIdsByLine(line) {
        const result = [];
        if (this.rangeOwner.size === 1) {
            // 只有一个provider，直接返回
            return [this.rangeOwner.keys().next().value];
        }
        for (const rangeOwner of this.rangeOwner) {
            const [id, ranges] = rangeOwner;
            if (ranges && ranges.some((range) => range.startLineNumber <= line && line <= range.endLineNumber)) {
                result.push(id);
            }
        }
        return result;
    }
};
tslib_1.__decorate([
    (0, di_1.Autowired)(di_1.INJECTOR_TOKEN),
    tslib_1.__metadata("design:type", di_1.Injector)
], CommentsService.prototype, "injector", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_browser_1.AppConfig),
    tslib_1.__metadata("design:type", Object)
], CommentsService.prototype, "appConfig", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(browser_1.IEditorDecorationCollectionService),
    tslib_1.__metadata("design:type", Object)
], CommentsService.prototype, "editorDecorationCollectionService", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_theme_1.IIconService),
    tslib_1.__metadata("design:type", Object)
], CommentsService.prototype, "iconService", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_browser_1.LabelService),
    tslib_1.__metadata("design:type", ide_core_browser_1.LabelService)
], CommentsService.prototype, "labelService", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(common_1.ICommentsFeatureRegistry),
    tslib_1.__metadata("design:type", Object)
], CommentsService.prototype, "commentsFeatureRegistry", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(browser_1.IEditorDocumentModelService),
    tslib_1.__metadata("design:type", Object)
], CommentsService.prototype, "documentService", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_main_layout_1.IMainLayoutService),
    tslib_1.__metadata("design:type", Object)
], CommentsService.prototype, "layoutService", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(browser_1.WorkbenchEditorService),
    tslib_1.__metadata("design:type", browser_1.WorkbenchEditorService)
], CommentsService.prototype, "workbenchEditorService", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(browser_1.ResourceService),
    tslib_1.__metadata("design:type", browser_1.ResourceService)
], CommentsService.prototype, "resourceService", void 0);
tslib_1.__decorate([
    ide_core_browser_1.memoize,
    tslib_1.__metadata("design:type", Object),
    tslib_1.__metadata("design:paramtypes", [])
], CommentsService.prototype, "isMultiCommentsForSingleLine", null);
tslib_1.__decorate([
    ide_core_browser_1.memoize,
    tslib_1.__metadata("design:type", Object),
    tslib_1.__metadata("design:paramtypes", [])
], CommentsService.prototype, "currentAuthorAvatar", null);
tslib_1.__decorate([
    ide_core_browser_1.memoize,
    tslib_1.__metadata("design:type", Object),
    tslib_1.__metadata("design:paramtypes", [])
], CommentsService.prototype, "filterThreadDecoration", null);
CommentsService = tslib_1.__decorate([
    (0, di_1.Injectable)()
], CommentsService);
exports.CommentsService = CommentsService;
//# sourceMappingURL=comments.service.js.map